==> entity_query_abm-2.0.72259a0.tar.gz: entity_query_abm-2.0.72259a0/entity_query/reference.py <==
#  Copyright (C) 2023
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import functools
import gc
import sys
from dataclasses import dataclass, field
from typing import Generic, Self, ForwardRef

import sqlalchemy.orm
from entity_read.entity import Entity
from entity_read.utils.sqlalchemy_dataclass_entity import default_registry
from more_itertools import first

from .column import Column, ColumnType
from .expression import Expression
from .order import Order
from .relation import EntityType
from .remote_entity import RemoteEntity


@dataclass(frozen=True, kw_only=True, repr=False)
class Reference(Column, Generic[ColumnType, EntityType]):
    key: str
    parent: RemoteEntity | type[RemoteEntity] | None = field(hash=False, default=None)

    def __class_getitem__(cls, params):
        column_type = params[0]
        entity_type = params[1] if len(params) > 1 else None
        if entity_type and hasattr(entity_type, "__name__"):
            name = f"{cls.__name__}[{column_type.__name__},{entity_type.__name__}]"
        else:
            name = f"{cls.__name__}[{column_type}]"

        attrs = {"_column_type": column_type}
        if entity_type:
            attrs["_entity_type"] = entity_type

        return type(name, (cls,), attrs)

    def __call__(self, parent: RemoteEntity | type[RemoteEntity] | None) -> "Reference[ColumnType, EntityType]":
        return type(self)(key=self.key, parent=parent)

    def shortcut_over(self) -> str:
        return f"ref.{self.key}"

    def base(self) -> Self:
        return Reference(key=self.key, parent=None)

    @functools.cached_property
    def _type_arg(self) -> type[EntityType]:
        entity_type = getattr(self, "_entity_type")
        if isinstance(entity_type, str):
            entity_type = self.__class__.__globals__[entity_type]
        return entity_type

    @functools.cached_property
    def related(self) -> EntityType:
        return self._type_arg(parent=self)

    def get_related_entity(self, entity_type: type[Entity]) -> type[Entity]:
        referenced_column = self._get_referenced_column(entity_type)
        referenced_table = referenced_column.table
        referenced_entity = get_entity_by_table(referenced_table)
        return referenced_entity

    def get_remote_column_name_to_local(self, entity_type: type[Entity]) -> dict[str, str]:
        referenced_column = self._get_referenced_column(entity_type)
        return {referenced_column.name: self.key}

    def _get_referenced_column(self, entity_type: type[Entity]) -> sqlalchemy.Column:
        if (column := entity_type.get_key_to_column().get(self.key)) is None:
            raise KeyError(f"Not found column {self.key!r} in {entity_type.__name__}({entity_type.get_column_names()})")
        foreign_keys = column.foreign_keys
        if not foreign_keys:
            raise KeyError(f"Column {self.key!r} in {entity_type.__name__} references nothing")
        if len(foreign_keys) > 1:
            raise NotImplementedError(f"Column {self.key!r} in {entity_type.__name__} references multiple tables")
        foreign_key: sqlalchemy.ForeignKey = first(foreign_keys)
        referenced_column = foreign_key.column
        return referenced_column

    def subquery(
        self,
        attrs: list[Expression] | None = None,
        vars: dict[str, Expression] | None = None,
        filters: list[Expression] | None = None,
        searches: list[Expression] | None = None,
        orders: list[Order] | None = None,
        limit: int | None = None,
        offset: int | None = None,
    ) -> 'SubQuery':
        from .subquery import SubQuery
        return SubQuery(
            over=self, attrs=attrs or [], vars=vars or {}, filters=filters or [], searches=searches or [],
            orders=orders or [], limit=limit, offset=offset
        )

    def exists(
        self,
        attrs: list[Expression] | Expression | None = None,
        vars: dict[str, Expression] | None = None,
        filters: list[Expression] | None = None,
        searches: list[Expression] | None = None,
        orders: list[Order] | None = None,
        limit: int | None = None,
        offset: int | None = None,
    ) -> 'Exists':
        from .function import Exists
        return Exists(self.subquery(
            attrs=attrs, vars=vars, filters=filters, searches=searches, orders=orders, limit=limit, offset=offset
        ))


def get_entity_by_table(
        table: sqlalchemy.Table,
        registry: sqlalchemy.orm.registry = default_registry
) -> type[Entity] | None:
    found_classes = {
        cls for cls in registry._class_registry.values()
        if getattr(cls, '__table__', None) is table and issubclass(cls, Entity)
    }
    if len(found_classes) > 1:
        raise ValueError(f"Multiple classes found for table {table.name}.")
    elif found_classes:
        return first(found_classes)
    return None

==> entity_query_abm-2.0.72260a0.tar.gz: entity_query_abm-2.0.72260a0/entity_query/reference.py <==
#  Copyright (C) 2023
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import functools
import importlib
from dataclasses import dataclass, field
from typing import Generic, Self

import sqlalchemy.orm
from entity_read.entity import Entity
from entity_read.utils.sqlalchemy_dataclass_entity import default_registry
from more_itertools import first

from .column import Column, ColumnType
from .expression import Expression
from .order import Order
from .relation import EntityType
from .remote_entity import RemoteEntity


@dataclass(frozen=True, kw_only=True, repr=False)
class Reference(Column, Generic[ColumnType, EntityType]):
    key: str
    parent: RemoteEntity | type[RemoteEntity] | None = field(hash=False, default=None)

    def __class_getitem__(cls, params):
        column_type = params[0]
        entity_type = params[1] if len(params) > 1 else None
        if entity_type and hasattr(entity_type, "__name__"):
            name = f"{cls.__name__}[{column_type.__name__}, {entity_type}]"
        else:
            name = f"{cls.__name__}[{column_type.__name__}]"

        attrs = {"_column_type": column_type}
        if entity_type:
            attrs["_entity_type"] = entity_type

        return type(name, (cls,), attrs)

    def __call__(self, parent: RemoteEntity | type[RemoteEntity] | None) -> "Reference[ColumnType, EntityType]":
        return type(self)(key=self.key, parent=parent)

    def shortcut_over(self) -> str:
        return f"ref.{self.key}"

    def base(self) -> Self:
        return Reference(key=self.key, parent=None)

    @functools.cached_property
    def _type_arg(self) -> type[EntityType]:
        entity_type = getattr(self, "_entity_type")
        if isinstance(entity_type, str):
            parent = self.parent
            module = importlib.import_module(parent.__class__.__module__)
            entity_type = getattr(module, entity_type)
            object.__setattr__(self, "_entity_type", entity_type)
        return entity_type

    @functools.cached_property
    def related(self) -> EntityType:
        return self._type_arg(parent=self)

    def get_related_entity(self, entity_type: type[Entity]) -> type[Entity]:
        referenced_column = self._get_referenced_column(entity_type)
        referenced_table = referenced_column.table
        referenced_entity = get_entity_by_table(referenced_table)
        return referenced_entity

    def get_remote_column_name_to_local(self, entity_type: type[Entity]) -> dict[str, str]:
        referenced_column = self._get_referenced_column(entity_type)
        return {referenced_column.name: self.key}

    def _get_referenced_column(self, entity_type: type[Entity]) -> sqlalchemy.Column:
        if (column := entity_type.get_key_to_column().get(self.key)) is None:
            raise KeyError(f"Not found column {self.key!r} in {entity_type.__name__}({entity_type.get_column_names()})")
        foreign_keys = column.foreign_keys
        if not foreign_keys:
            raise KeyError(f"Column {self.key!r} in {entity_type.__name__} references nothing")
        if len(foreign_keys) > 1:
            raise NotImplementedError(f"Column {self.key!r} in {entity_type.__name__} references multiple tables")
        foreign_key: sqlalchemy.ForeignKey = first(foreign_keys)
        referenced_column = foreign_key.column
        return referenced_column

    def subquery(
        self,
        attrs: list[Expression] | None = None,
        vars: dict[str, Expression] | None = None,
        filters: list[Expression] | None = None,
        searches: list[Expression] | None = None,
        orders: list[Order] | None = None,
        limit: int | None = None,
        offset: int | None = None,
    ) -> 'SubQuery':
        from .subquery import SubQuery
        return SubQuery(
            over=self, attrs=attrs or [], vars=vars or {}, filters=filters or [], searches=searches or [],
            orders=orders or [], limit=limit, offset=offset
        )

    def exists(
        self,
        attrs: list[Expression] | Expression | None = None,
        vars: dict[str, Expression] | None = None,
        filters: list[Expression] | None = None,
        searches: list[Expression] | None = None,
        orders: list[Order] | None = None,
        limit: int | None = None,
        offset: int | None = None,
    ) -> 'Exists':
        from .function import Exists
        return Exists(self.subquery(
            attrs=attrs, vars=vars, filters=filters, searches=searches, orders=orders, limit=limit, offset=offset
        ))


def get_entity_by_table(
        table: sqlalchemy.Table,
        registry: sqlalchemy.orm.registry = default_registry
) -> type[Entity] | None:
    found_classes = {
        cls for cls in registry._class_registry.values()
        if getattr(cls, '__table__', None) is table and issubclass(cls, Entity)
    }
    if len(found_classes) > 1:
        raise ValueError(f"Multiple classes found for table {table.name}.")
    elif found_classes:
        return first(found_classes)
    return None
